from multiprocessing import Process, Value
from ctypes import c_longlong, c_byte
import os
import signal
import sys
import time

from bloonix.logger import logger
from bloonix.json import json

class ProcManager:
    def __init__(self, signal_hup_callback=None, signal_term_callback=None):
        self.pools = []
        self.signal_hup_callback = signal_hup_callback
        self.signal_term_callback = signal_term_callback
        self.done = False
        self.log = logger("bloonix")
        self.daemon_callback = None

    def add_pool(
            self, name, worker, callback,
            exit_on_signal_term=True,
            signal_hup_callback=None,
            signal_term_callback=None,
            args=[], kwargs={}):
        self.log.notice("register pool %s with %s worker" % (name, worker))
        if worker > 0:
            pool = Pool(
                name=name,
                worker=worker,
                callback=callback,
                exit_on_signal_term=exit_on_signal_term,
                signal_hup_callback=signal_hup_callback,
                signal_term_callback=signal_term_callback,
                args=args,
                kwargs=kwargs)
            self.pools.append(pool)

    def add_daemon_callback(self, callback):
        """
        This callback is called on each loop within daemonize()
        and can be used to debug things, like
            print( p.process_status() )
        """
        self.daemon_callback = callback

    def daemonize(self):
        self.log.info("start main proc manager")
        signal.signal(signal.SIGHUP,  self._sig_hup_handler)
        signal.signal(signal.SIGTERM, self._sig_term_handler)
        signal.signal(signal.SIGINT, self._sig_term_handler)

        while self.done is False:
            self._check_pools()

        self._kill_pools()
        self.log.info("stop proc manager")

    def _check_pools(self):
        callback = self.daemon_callback
        while self.done is False:
            for pool in self.pools:
                self._check_pool(pool)

            if callback: callback()
            # sleep only if done is False, so dieing as fast
            # as possilbe on sig{int|term}
            if self.done is False:
                time.sleep(1)

    def _check_pool(self, pool):
        self._spawn_worker(pool)
        self._reap_worker(pool)

    def _spawn_worker(self, pool):
        num = len(pool.children)

        for i in range(pool.worker - num):
            if not len(pool.status_objects) or not len(pool.request_objects):
                self.log.warning("not enough value objects in pool")
                return
            self.log.notice("start worker [pool=%s]" % pool.name)
            # Locking is not necessary because the forked worker
            # is the only one who ever writes to the shared memory.
            status = pool.status_objects.pop()
            request = pool.request_objects.pop()
            status.value = 1
            worker = Worker(
                pool=pool,
                status=status,
                request=request,
                log=self.log,
                process_status=self.process_status,
                process_size=self.process_size,
            )
            p = Process(target=worker._run)
            p.start()
            pool.children[p.pid] = Child(p, status, request)

    def _reap_worker(self, pool):
        for pid in list(pool.children):
            child = pool.children[pid]
            if not child.process.is_alive():
                child.status.value = 0
                pool.status_objects.append(child.status)
                pool.request_objects.append(child.request)
                child.process.join()
                del pool.children[pid]
                if child.process.exitcode > 0:
                    self.log.error(
                        "child pid %s died with exitcode %s [pool=%s]"
                        % (pid, child.process.exitcode, pool.name))

    def _kill_pools(self):
        children = {}

        for pool in self.pools:
            for pid in list(pool.children):
                child = pool.children[pid]
                if child.process.is_alive():
                    self.log.info("send sigterm to pid %s [pool=%s]" % (
                        pid, pool.name))
                    try:
                        os.kill(pid, 15)
                    except Exception:
                        pass
                    children[pid] = [child, pool.name]
                    del pool.children[pid]

        timeout = time.time() + 15

        while len(children) > 0 and timeout > time.time():
            self.log.info("children still alive:", list(children))
            for pid in list(children):
                child, pool_name = children[pid]
                if not child.process.is_alive():
                    child.process.join()
                    del children[pid]

            time.sleep(0.1)

        for pid in list(children):
            child, pool_name = children[pid]
            self.log.info("send sigkill to pid %s [pool=%s]" % (pid, pool_name))
            try:
                os.kill(pid, 9)
            except Exception:
                pass

    def process_status(self, simple=None):
        if simple is True:
            return self._get_simple_process_status()
        data = {}
        for pool in self.pools:
            data[pool.name] = {
                "requests": 0,
                "initiazing": 0,
                "waiting": 0,
                "reading": 0,
                "processing": 0,
                "sending": 0,
                "total": 0,
            }
            for o in pool.all_status_objects:
                if   o.value == 1: data[pool.name]["initiazing"] += 1
                elif o.value == 2: data[pool.name]["waiting"] += 1
                elif o.value == 3: data[pool.name]["reading"] += 1
                elif o.value == 4: data[pool.name]["processing"] += 1
                elif o.value == 5: data[pool.name]["sending"] += 1
                data[pool.name]["total"] += 1
            for o in pool.all_request_objects:
                data[pool.name]["requests"] += o.value
        return data

    def _get_simple_process_status(self):
        data = {}
        for pool in self.pools:
            n = pool.name
            data["%s_initiazing" % n] = 0
            data["%s_waiting" % n] = 0
            data["%s_reading" % n] = 0
            data["%s_processing" % n] = 0
            data["%s_sending" % n] = 0
            data["%s_total" % n] = 0
            data["%s_requests" % n] = 0

        for pool in self.pools:
            n = pool.name
            for o in pool.all_status_objects:
                if   o.value == 1: data["%s_initiazing" % n] += 1
                elif o.value == 2: data["%s_waiting" % n] += 1
                elif o.value == 3: data["%s_reading" % n] += 1
                elif o.value == 4: data["%s_processing" % n] += 1
                elif o.value == 5: data["%s_sending" % n] += 1
                data["%s_total" % n] += 1
            for o in pool.all_request_objects:
                data["%s_requests" % n] += o.value
        return data

    def process_size(self):
        return Size()

    def send_signal_to_pool(self, name=None, signal=1):
       for pool in self.pools:
            if name is not None and name != pool.name:
                continue
            for pid in list(pool.children):
                child = pool.children[pid]
                if child.process.is_alive():
                    try:
                        os.kill(pid, signal)
                    except Exception:
                        pass

    def _sig_term_handler(self, signum, frame):
        if self.signal_term_callback:
            self.signal_term_callback()
        self.log.info("signal TERM/INT received - quitting")
        self.done = True

    def _sig_hup_handler(self, signum, frame):
        self.log.info("signal HUP received")
        if self.signal_hup_callback:
            self.signal_hup_callback()


class Worker:
    def __init__(
            self, pool, status, request, log,
            process_status, process_size):
        self.pool = pool
        self.status = status
        self.request = request
        self.log = log
        self.process_status = process_status
        self.process_size = process_size
        self.exit_on_signal_term = pool.exit_on_signal_term
        self.signal_hup_callback = pool.signal_hup_callback
        self.signal_term_callback = pool.signal_term_callback
        self.done = False

    def is_waiting(self):
        self.status.value = 2

    def is_reading(self):
        self.status.value = 3

    def is_processing(self):
        self.status.value = 4
        self.request.value += 1

    def is_sending(self):
        self.status.value = 5

    def _sig_hup_handler(self, signum, frame):
        self.log.info("signal HUP received [pool=%s]" % self.pool.name)
        if self.signal_hup_callback:
            try: self.signal_hup_callback()
            except Exception: self.log.exception()

    def _sig_term_handler(self, signum, frame):
        self.log.info("signal TERM/INT received - quitting [pool=%s]" % self.pool.name)
        if self.signal_term_callback:
            try: self.signal_term_callback()
            except Exception: self.log.exception()
        if self.exit_on_signal_term:
            sys.exit(0)
        self.done = True

    def set_signal_hup_callback(self, callback):
        self.signal_hup_callback = callback

    def set_signal_term_callback(self, callback):
        self.signal_term_callback = callback

    def _run(self):
        signal.signal(signal.SIGHUP, self._sig_hup_handler)
        signal.signal(signal.SIGTERM, self._sig_term_handler)
        signal.signal(signal.SIGINT, self._sig_term_handler)
        try:
            self.pool.callback(self, *self.pool.args, **self.pool.kwargs)
        except Exception:
            self.log.exception()
        sys.exit(0)


class Pool:
    def __init__(
            self, name, worker, callback,
            exit_on_signal_term,
            signal_hup_callback,
            signal_term_callback,
            args, kwargs):
        self.name = name
        self.worker = worker
        self.callback = callback
        self.exit_on_signal_term = exit_on_signal_term
        self.signal_hup_callback = signal_hup_callback
        self.signal_term_callback = signal_term_callback
        self.args = args
        self.kwargs = kwargs
        self.children = {}
        self.status_objects = []
        self.request_objects = []
        for i in range(worker):
            self.status_objects.append(
                Value(c_byte, 0, lock=False))
            self.request_objects.append(
                Value(c_longlong, 0, lock=False))
        self.all_status_objects = list(self.status_objects)
        self.all_request_objects = list(self.request_objects)


class Child:
    def __init__(self, process, status, request):
        self.process = process
        self.status = status
        self.request = request


class Size:
    def __init__(self):
        self.size = 0
        self.resident = 0
        self.share = 0
        self.trs = 0
        self.lrs = 0
        self.drs = 0
        self.dtp = 0
        filename = "/proc/%s/statm" % os.getpid()
        try:
            with open(filename, encoding="utf-8") as f:
                line = f.readline()
                f.close()
                parts = list(map((lambda x: int(x) * 4096), line.split()))
                self.size = parts[0]
                self.resident = parts[1]
                self.shared = parts[2]
                self.text = parts[3]
                self.lib = parts[4]
                self.data = parts[5]
                self.dt = parts[6]
        except Exception:
            pass
